/*
 * $Id: corpwl.c,v 1.1 1996/01/30 20:47:53 scotty Exp scotty $
 *
 * $Log: corpwl.c,v $
 * Revision 1.1  1996/01/30 20:47:53  scotty
 * Initial revision
 *
 */
#include "corpinc.h"

static char rcsid[] = "$Id: corpwl.c,v 1.1 1996/01/30 20:47:53 scotty Exp scotty $";

#define SERVER 1

#define NWL 10

struct {
   char      pattern[MAXWORD];
   char      path[_MAX_PATH];
   int       count;
} wl[NWL];

BOOL      HashWordList(unsigned, int, struct regexp *);
BOOL      bMatch(char *, char *);
BOOL      MergeWord(char *, int);
BOOL      MergeToDisk(struct stream *, struct stream *, FILE *);
BOOL      GetWLName(char *, char *);
void      DelLastWL(void);
void      GetSet(BOOL[256], const char *, int *);

struct regexp *ParseRE(const char *, int *);
struct regexp *ParseDJ(const char *, int *);
struct regexp *ParseCJ(const char *, int *);
struct regexp *ParseCC(const char *, int *);
struct regexp *ParseRegExp(const char *);

BOOL      pMatch(const char *, struct regexp *);
BOOL      pMatch0(char **, struct regexp *);
BOOL      pMatch1(char **, struct regexp *);
BOOL      pMatchHash(unsigned, struct regexp *, int *);
BOOL      pMatchHash1(unsigned, struct regexp *, int *);
BOOL      IString(struct regexp *, char *);
BOOL      IString1(struct regexp *, char *);

char      tn[_MAX_PATH];
char      pn[_MAX_PATH];

int       nw,
          nq;
int       thisi;

#if !defined(SERVER)

void WordList(void)
{
   int            i;
   FILE          *f;
   struct regexp *re;
   char           patt[MAXWORD];

   strcpy(tn, tmppath);
   strcat(tn, "wctemp");

   EditGetWord("Enter pattern: ", patt);
   re = ParseRegExp(patt);
   strcpy(qs.xword, patt);
   preproc(patt);

   qs.type = TY_FILE;
   qs.first = 0;
   qs.ff = 0;

   if (GetWLName(patt, pn))
      return;

   f = fopen(pn, "w");
   fclose(f);

   nw = nq = 0;

   for (i = 0; i < HASHRANGE; i++)
      HashWordList(i, 0, re);

   printf("%i words found\n%i solutions\n", nw, nq);

   qs.last = nq % MAXILOC;
   qs.lf = nq / MAXILOC;
   qs.count = wl[thisi].count = nq;
   qs.handle = fopen(pn, "r");

   return;
}

#endif /* ! SERVER */

int WoL(char *patt)
{
   int            i;
   FILE          *f;
   struct regexp *re;

   ScratchFile(tn);

   re = ParseRegExp(patt);   
   if (re==NULL) return 0; /* should give an error as this means 'can't parse' */
   strcpy(qs.xword, patt);
   preproc(patt);

   qs.type = TY_FILE;
   qs.first = 0;
   qs.ff = 0;

   f = fopen(pn, "w");
   fclose(f);

   nw = nq = 0;

   for (i = 0; i < hashrange; i++)
      if (!HashWordList(i, 0, re)) {remove(pn); return 0;} 

   return nq;
}


BOOL HashWordList(unsigned hi, int k, struct regexp *re)
{                    
	struct ihdr header;
   char      path[_MAX_PATH];
   FILE     *fInd;
   int       i;
   int       ji = 0;
   off_t     loff;

   if (!pMatchHash(hi, re, &ji))
      return TRUE;                                 
loop:      
   IndexFile(hi, k, path);
/*
	if (k==0)
      sprintf(path,"%s%i.HXX",IndexDir(hi),hi);
	else
      sprintf(path,"%s%i.H%iX",IndexDir(hi),hi,k);
*/
   fInd = fopen(path, "r");
   if (fInd == NULL)
      return TRUE;

   fread(&header, sizeof(header), 1, fInd);
   fix_ihdr(&header, 1);
   if (header.version != INDEXVER) {
      corperr("Index %s wrong format", path);
      fclose(fInd);
      return FALSE;
   }
   fread(dict, sizeof(struct dicent), header.ndict, fInd);
   fix_dicent(dict, header.ndict);

   loff = sizeof(header) + header.ndict * sizeof(struct dicent)
      + header.nill * sizeof(struct locent);
   fseek(fInd, loff, SEEK_SET);
   fread(words, 1, header.nbytes, fInd);

   fclose(fInd);

   for (i = 0; i < header.ndict; i++) {
      char *w;
      w = words + dict[i].dwd;
      if (pMatch(w, re)  && dict[i].type == ID_WORD && (bNoPos || dict[i].extra != HDRPOS))
         if (!MergeWord(words + dict[i].dwd, dict[i].extra)) return FALSE;
	  TestInterrupt();
   }

   if (header.bOS)
      {k++; goto loop;}

   return TRUE;
}

BOOL MergeWord(char *word, int extra)
{
   struct stream *s1;
   FILE          *ft,
                 *fp;
	BOOL b;                 

#if !defined(SERVER)
   if (!bServer)
      printf("%s\n", word);
#endif /* ! SERVER */

   nw++;

   s1 = MakeAtomStream(word, extra, ID_WORD);

   ft = fopen(tn, "w");
   fp = fopen(pn, "r");

   qs.handle = fp;
   qs.lf = nq / maxiloc;
   qs.last = nq % maxiloc;
   qs.count = nq;

   b=MergeToDisk(s1, &qs, ft);
   KillFileBuffers(fp);

   fclose(fp);
   fclose(ft);

   remove(pn);
   rename(tn, pn);

   FreeStream(s1);

   return b;
}


BOOL bMatch(char *word, char *pattern)
{
   int       i = 0;
   int       j = 0;
   BOOL      cs[256];

   while (pattern[i] != 0) {
      switch (pattern[i]) {
       case '.':
         j++;
         break;

       case '*':
         if (pattern[i + 1] == 0)
            return TRUE;

         while (word[j] != 0)
            if (bMatch(word + j++, pattern + i + 1))
               return TRUE;

         return FALSE;

       case '[':
         GetSet(cs, pattern, &i);
         if (!cs[word[j]])
            return FALSE;

         j++;
         break;

       case '\\':
         i++;
         /* FALLTHROUGH */
         
       default:
         if (word[j] != pattern[i])
            return FALSE;

         j++;

      }
      i++;
   }

   return word[j] == 0;
}


void GetSet(BOOL b[256], const char *p, int *i)
{
   BOOL      bNeg = FALSE;
   char      c1,
             c2;
   int       j;

   (*i)++;

   if (p[*i] == '^') {
      bNeg = TRUE;
      (*i)++;
   }

   for (j = 0; j < 256; j++)
      b[j] = bNeg;

   c1 = p[*i];
   (*i)++;

xy:
   c2 = p[*i];
   (*i)++;

   switch (c2) {
    case ']':
      b[c1] = !bNeg;
      return;

    case '-':
      c2 = p[*i];
      (*i)++;
      if (c2 == ']') {
         b[c1] = !bNeg;
         return;
      }
      for (j = c1; j < c2; j++)
         b[j] = !bNeg;

      break;

    default:
      b[c1] = !bNeg;
      break;
   }

   c1 = c2;

   goto xy;
}


BOOL GetWLName(char *pattern, char *path)
{
   int   i;

   for (i = 0; i < NWL; i++)
      if (!strcmp(pattern, wl[i].pattern))
         break;

   if (i < NWL) {
      strcpy(path, wl[i].path);
      return TRUE;
   }
   for (i = 0; i < NWL; i++)
      if (wl[i].pattern[0] == 0)
         break;

   if (i == NWL) {
      remove(wl[0].path);
      i = 0;
   }
   strcpy(wl[i].pattern, pattern);

   sprintf(path, "%s%i.wl", tmppath, i);
   strcpy(wl[i].path, path);

   thisi = i;
   return FALSE;
}

void DelLastWL(void)
{
   remove(wl[thisi].path);
   wl[thisi].path[0] = 0;
   return;
}

void DelAllWL(void)
{
   int       i;

   for (i = 0; i < NWL; i++)
      if (wl[i].path[0] != 0)
	 remove(wl[i].path);

   return;
}


BOOL MergeToDisk(struct stream *s1, struct stream *s2, FILE *f)
{
   BOOL           b1,
                  b2;
   WORDNO         w1,
                  w2;
   struct locent1 l;
   sara_ulong     c1,
                  c2;

   nq = 0;

   RewindStream(s1);
   RewindStream(s2);

   b1 = AdvanceStream(s1);
   b2 = AdvanceStream(s2);

   while (b1 || b2) {
      w1 = s1->wpos;
      c1 = s1->lcode;

      w2 = s2->wpos;
      c2 = s2->lcode;
      if (!b1 || (b2 && (c2 < c1 || (c1 == c2 && w2 < w1)))) {
         l.pos = s2->pos;
         l.word = w2;
         l.lcode = s2->lcode;
         l.lw = s2->lw;
		 l.lw0 = s2->lw0;
         if (fwrite(&l, sizeof(l), 1, f) < 1) return FALSE;

         b2 = AdvanceStream(s2);
      } else {
         l.pos = s1->pos;
         l.word = w1;
         l.lcode = s1->lcode;
         l.lw = s1->lw;
		 l.lw0 = s1->lw0;
         if (fwrite(&l, sizeof(l), 1, f) < 1) return FALSE;

         b1 = AdvanceStream(s1);
      }
      nq++;
   }

   return TRUE;
}


#define NOPS 100
#define NSET 4

struct regexp  op[NOPS];

BOOL           cs[NSET][256];

int            ncell,
               nset;


struct regexp *ParseRegExp(const char *pattern)
{
   int   i = 0;

   ncell = nset = 0;
   return ParseRE(pattern, &i);
}



struct regexp *ParseRE(const char *pattern, int *i)
{
   struct regexp *op1 = ParseDJ(pattern, i);

   while (TRUE) {
      switch (pattern[*i]) {
       case 0:
       case ')':
         return op1;

      case '|':
         op[ncell].op = '|';
         op[ncell].op1 = op1;
         op[ncell].rep = '1';
         op1 = &op[ncell++];
         (*i)++;

         op1->op2 = ParseDJ(pattern, i);
         break;
      }        
      if (ncell >= NOPS) return NULL; 
   }
}


struct regexp *ParseDJ(const char *pattern, int *i)
{
   struct regexp *op1 = ParseCJ(pattern, i);

   while (pattern[*i] != 0 && pattern[*i] != ')' && pattern[*i] != '|') {
      op[ncell].op = '+';
      op[ncell].op1 = op1;
      op[ncell].rep = '1';
      op1 = &op[ncell++];
      op1->op2 = ParseCJ(pattern, i);
   }

   return op1;
}


struct regexp *ParseCJ(const char *pattern, int *i)
{
   struct regexp *op1 = ParseCC(pattern, i);

   switch (pattern[*i]) {
    case '*':
    case '?':
      op1->rep = pattern[*i];
      (*i)++;
      break;
    case '+':            
    	(*i)++;
    	op1->rep='1';
      op[ncell]=*op1;
      op[ncell].rep='*';
      ncell++;
      op[ncell].op = '+';              
      op[ncell].rep='1';
      op[ncell].op1 = op1;
      op[ncell].op2 = &op[ncell-1];
      return &op[ncell++];
   }

   return op1;
}


struct regexp *ParseCC(const char *pattern, int *i)
{
   struct regexp *op1;

   switch (pattern[*i]) {
    case '(':
      (*i)++;
      op1 = ParseRE(pattern, i);
      if (pattern[*i] == ')')
         (*i)++;
      return op1;

   case '[':
      GetSet(cs[nset], pattern, i);
      op[ncell].op = 's';
      op[ncell].set = nset++;
      op[ncell].rep = '1';
      op1 = &op[ncell++];
      return op1;

   case '.':
      (*i)++;
      op[ncell].op = 'a';
      op[ncell].rep = '1';
      op1 = &op[ncell++];
      return op1;

   case '\\':
      (*i)++;

   default:
      op[ncell].op = 'c';
      op[ncell].ch = pattern[*i];
      op[ncell].rep = '1';
      op1 = &op[ncell++];
      (*i)++;
      return op1;
   }
}

BOOL IString(struct regexp *re, char *s)
{
   switch (re->rep) {
    case '1':
      return IString1(re, s);

    case '*':
      s[0] = 0;
      return FALSE;

    case '+':
      return IString1(re, s);

    case '?':
      s[0] = 0;
      return FALSE;
   }
	return TRUE;
}


BOOL IString1(struct regexp *re, char *s)
{
   int       n1,
             i;
   char      s1[MAXWORD];

   s[0] = 0;
   switch (re->op) {
    case 'a':
      return FALSE;

    case 'c':
      s[0] = re->ch;
      s[1] = 0;
      return TRUE;

    case 's':
      return FALSE;

    case '+':
      if (IString(re->op1, s))
         return IString(re->op2, s + strlen(s));
      else
         return FALSE;

    case '|':
      IString(re->op1, s);
      IString(re->op2, s1);
      n1 = strlen(s);
      for (i = 0; i < n1; i++)
         if (s[i] != s1[i] || s1[i] == 0)
            break;
      s[i] = 0;
      return FALSE;
   }
	return TRUE;
}


BOOL pMatchHash(unsigned h, struct regexp *re, int *l)
{
   switch (re->rep) {
    case '1':
      return pMatchHash1(h, re, l);

    case '*':
    case '?':
	  *l=nhashsig+1;
      return TRUE;

    case '+':
      return TRUE;  


   }
	return TRUE;
}


BOOL pMatchHash1(unsigned h, struct regexp *re, int *l)
{
   switch (re->op) {
    case 'a':
      (*l)++;
      return TRUE;

    case 'c':
      return bMayHashChar(h, re->ch, (*l)++);

    case 's':
      (*l)++;
      return TRUE;			 /* lazy! */

   case '+':
      if (!pMatchHash(h, re->op1, l))
         return FALSE;

      return pMatchHash(h, re->op2, l);

   case '|':
      if (!pMatchHash(h,re->op1,l) && !pMatchHash(h,re->op2,l)) return FALSE;
      *l = 100;
      return TRUE;

   }
	return TRUE;
}


#define MAXALT 1000

struct {
	const char* w;
	struct regexp* e;      
	int cp;
} astack[MAXALT];	        


struct regexp *cstack[MAXALT];

int altp;                                             
int contp;

BOOL popcont(struct regexp** e)
{
	if (contp==0) return FALSE;
	contp--;
	*e=cstack[contp];
	return TRUE;
}	

BOOL popalt(const char** w,struct regexp** e)
{         
	if (altp==0) return FALSE;
	altp--;
	*w=astack[altp].w;
	*e=astack[altp].e;        
	contp=astack[altp].cp;
	return TRUE;
}

void pushalt(const char* w,struct regexp* e)
{
	if (altp>=MAXALT-1) {
		syslog(0,"Regexp stack A overflow");
		return;
	}
	astack[altp].w=w;
	astack[altp].e=e;
	astack[altp].cp=contp;
	altp++;
}		             

void pushcont(struct regexp* e)
{
	if (contp>=MAXALT-1) {
		syslog(0,"Regexp stack C overflow");
		return;
	}
	cstack[contp++]=e;  
}	

BOOL TryMatch();
BOOL OneMatch(char**);
	
BOOL pMatch(const char* word,struct regexp *re)
{                           
	altp=0;            
	contp=0;	
	pushalt(word,re);        
	return TryMatch();
}                         

BOOL TryMatch()
{              
	char * word; struct regexp* re;
	while (popalt((const char**) &word,&re)) {
		pushcont(re);
		if (OneMatch((char**) &word) && *word==0)  return TRUE;
	}	
	return FALSE;
}		 
		

BOOL OneMatch(char** word)
{                                         
	struct regexp *re;
	while (popcont(&re)) {
		if (!pMatch0(word,re)) return FALSE;
	}
	return TRUE;	
}	
	                   
BOOL pMatch0(char** word,struct regexp *re)
{	                          
	if (re==NULL) return TRUE;
	switch (re->rep)
	{
	case '1':     
		return pMatch1(word,re);
	case '*':      
		pushalt(*word,NULL);
		pushcont(re);
		return pMatch1(word,re);
/*
	case '+': 
		pushcont
		if (!pMatch1(word,re)) return FALSE;
		while (pMatch1(word,re)) ;
		return TRUE;
*/
	case '?':                  
		pushalt(*word,NULL);
		return pMatch1(word,re);
	}
	return TRUE;
}

BOOL pMatch1(char** word,struct regexp *re)
{                                   
	char *w;
	switch (re->op)
	{
	case 'a':
		if (**word) {(*word)++; return TRUE;} else return FALSE;
	case 'c':
		if (**word==re->ch) {(*word)++; return TRUE;} else return FALSE;
	case 's':
		if (cs[re->set][(unsigned char) **word]) {(*word)++; return TRUE;} else return FALSE;
	case '+':
		pushcont(re->op2);
		return pMatch0(word,re->op1);
	case '|':
		w=*word;                    
		pushalt(w,re->op2);
		return pMatch0(word,re->op1);
	}
	return TRUE;
}			


struct stream *MakeRegExpStream(char *patt)
{
   int            n;
   FILE          *f;
   struct stream *s;
   sara_long      ss;

   if (!GetWLName(patt, pn)) {
      n = WoL(patt);
      if (n == 0)
	 return NULL;
   }

   f = fopen(pn, "r");
   s = GrabStream();
   s->type = TY_FILE;

   fseek(f, (off_t)0, SEEK_END);
   ss = ftell(f);
   rewind(f);

   s->count = (int) ss / sizeof(struct locent1);
   s->first = 0;
   s->ff = 0;
   s->last = (int) s->count % maxiloc;
   s->lf = (int) s->count / maxiloc;
   s->handle = f;
   s->wc = 1;

   return s;
}



int       rebuff[MAXRE];
int       nre;

int WoL1(char *regexp)
{
   int            i;
   int            k = 0;
   char           pre[MAXWORD];
   struct regexp *re;
   char          *wd;
   int            r;

   re = ParseRegExp(regexp);
   IString(re, pre);
   r = GetFirstDictEntry(pre);
   for (i = r; i < ndix; i++) {
      wd = GetNextDictEntry();
      if (wd == NULL)
         break;
      if (pMatch(wd, re)) {
         rebuff[k] = i;
         k++;
         if (k > MAXRE)
            return -2;
      }
   }
   nre = k;
   return k;
}


void WoL2(int k, char *wd, int swd, sara_long *freq, sara_long *forms)
{
   GetDictEntry(fDict, rebuff[k], wd, swd, freq,forms);
   return;
}
